#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#
#
"""
TMCC1 Constants
"""

from __future__ import annotations

from enum import unique
from typing import Dict, Tuple

from ..command_def import CommandDef, CommandDefEnum
from ..constants import CommandPrefix, CommandScope, CommandSyntax
from ..constants import RELATIVE_SPEED_MAP, OfficialRRSpeeds, DEFAULT_ADDRESS


class TMCC1Enum(CommandDefEnum):
    """
    Marker Interface for all TMCC1 enums
    """

    pass


"""
    TMCC1 Protocol Constants
"""
TMCC1_COMMAND_PREFIX: int = 0xFE


@unique
class TMCC1CommandIdentifier(CommandPrefix):
    ENGINE = 0b00000000  # first 2 bits significant
    TRAIN = 0b11001000  # first 5 bits significant
    SWITCH = 0b01000000  # first 2 bits significant
    ACC = 0b10000000  # first 2 bits significant
    ROUTE = 0b11010000  # first 4 bits significant
    HALT = 0xFFFF  # all 16 bits significant
    SYNC = 0xF0F0  # first 12 bits significant

    @classmethod
    def classify(cls, byte_data) -> CommandScope:
        fb_value = int(byte_data[0])
        # because of the varying number of bits, we need to
        # perform different tests
        if fb_value == TMCC1CommandIdentifier.HALT:
            return CommandScope.SYSTEM
        if fb_value & 0b11010000 == TMCC1CommandIdentifier.ROUTE:
            return CommandScope.ROUTE
        if fb_value & 0b11000000 == TMCC1CommandIdentifier.ENGINE:
            return CommandScope.ENGINE
        if fb_value & 0b11000000 == TMCC1CommandIdentifier.SWITCH:
            return CommandScope.SWITCH
        if fb_value & 0b11000000 == TMCC1CommandIdentifier.ACC:
            return CommandScope.ACC
        if fb_value & 0b11111000 == TMCC1CommandIdentifier.TRAIN:
            return CommandScope.TRAIN
        raise ValueError(f"Cannot classify TMCC identifier: {byte_data.hex(':')}")


TMCC1_IDENT_TO_SCOPE_MAP: Dict[TMCC1CommandIdentifier, CommandScope] = {
    TMCC1CommandIdentifier.ENGINE: CommandScope.ENGINE,
    TMCC1CommandIdentifier.TRAIN: CommandScope.TRAIN,
    TMCC1CommandIdentifier.SWITCH: CommandScope.SWITCH,
    TMCC1CommandIdentifier.ACC: CommandScope.ACC,
    TMCC1CommandIdentifier.ROUTE: CommandScope.ROUTE,
    TMCC1CommandIdentifier.HALT: CommandScope.SYSTEM,
    TMCC1CommandIdentifier.SYNC: CommandScope.SYNC,
}


class TMCC1CommandDef(CommandDef):
    def __init__(
        self,
        command_bits: int,
        command_ident: TMCC1CommandIdentifier = TMCC1CommandIdentifier.ENGINE,
        is_addressable: bool = True,
        num_address_bits: int = 7,
        d_min: int = 0,
        d_max: int = 0,
        d_map: Dict[int, int] = None,
        do_reverse_lookup: bool = True,
        alias: str = None,
        data: int = None,
        filtered: bool = False,
        aux1: bool = False,
        interval: int = None,
    ) -> None:
        super().__init__(
            command_bits,
            is_addressable=is_addressable,
            num_address_bits=num_address_bits,
            d_min=d_min,
            d_max=d_max,
            d_map=d_map,
            do_reverse_lookup=do_reverse_lookup,
            alias=alias,
            data=data,
            filtered=filtered,
            interval=interval,
        )
        self._command_ident = command_ident
        self._aux1 = aux1

    @property
    def syntax(self) -> CommandSyntax:
        return CommandSyntax.TMCC

    @property
    def identifier(self) -> TMCC1CommandIdentifier:
        return self._command_ident

    @property
    def first_byte(self) -> bytes:
        return TMCC1_COMMAND_PREFIX.to_bytes(1, byteorder="big")

    @property
    def scope(self) -> CommandScope:
        return TMCC1_IDENT_TO_SCOPE_MAP[self.identifier]

    @property
    def is_aux1_prefixed(self) -> bool:
        return self._aux1

    @property
    def address_mask(self) -> int:
        return 0xFFFF & ~((2**self.num_address_bits - 1) << 7)

    @property
    def train_address_mask(self) -> int:
        return 0xFFFF & ~(((2**4) - 1) << 7)

    def address_from_bytes(self, byte_data: bytes) -> int:
        if self.is_addressable:
            value = int.from_bytes(byte_data, byteorder="big")
            # if first byte indicates train, we have to change the address mask
            if self.classify(byte_data) == CommandScope.TRAIN:
                return (0xFFFF & (~self.train_address_mask & value)) >> 7
            else:
                return (0xFFFF & (~self.address_mask & value)) >> 7
        else:
            return DEFAULT_ADDRESS

    @staticmethod
    def classify(byte_data: bytes) -> CommandScope:
        return TMCC1CommandIdentifier.classify(byte_data)

    @property
    def alias(self) -> TMCC1EngineCommandEnum | Tuple[TMCC1EngineCommandEnum, int] | None:
        if self._alias is not None:
            if isinstance(self._alias, str):
                alias = TMCC1EngineCommandEnum.by_name(self._alias, raise_exception=True)
                if self._data is None:
                    return alias
                else:
                    return alias, self._data
            else:
                raise ValueError(f"Cannot classify command alias: {self._alias} ({type(self._alias)})")
        return None


class SyncCommandDef(TMCC1CommandDef):
    def __init__(
        self,
        command_bits: int,
        node_scope: bool = False,
        d_min: int = 0,
        d_max: int = 0,
        d_map: Dict[int, int] = None,
        do_reverse_lookup: bool = True,
        alias: str = None,
        data: int = None,
        filtered: bool = False,
    ) -> None:
        super().__init__(
            command_bits,
            command_ident=TMCC1CommandIdentifier.SYNC,
            is_addressable=False,
            num_address_bits=7,
            d_min=d_min,
            d_max=d_max,
            d_map=d_map,
            do_reverse_lookup=do_reverse_lookup,
            alias=alias,
            data=data,
            filtered=filtered,
        )
        self._node_scope = node_scope

    @property
    def is_node_scope(self) -> bool:
        return self._node_scope


TMCC1_REGISTER_COMMAND: int = 0xF0F0
TMCC1_DISCONNECT_COMMAND: int = 0xF0F1
TMCC1_SYNC_REQUEST_COMMAND: int = 0xF0F2
TMCC1_SYNC_BEGIN_COMMAND: int = 0xF0F3
TMCC1_SYNCING_COMMAND: int = 0xF0F4
TMCC1_SYNC_COMPLETE_COMMAND: int = 0xF0F5
TMCC1_SENDING_STATE_COMMAND: int = 0xF0F6
TMCC1_QUIT_COMMAND: int = 0xF0F7
TMCC1_REBOOT_COMMAND: int = 0xF0F8
TMCC1_UPDATE_COMMAND: int = 0xF0F9
TMCC1_UPGRADE_COMMAND: int = 0xF0FA
TMCC1_SHUTDOWN_COMMAND: int = 0xF0FB
TMCC1_RESTART_COMMAND: int = 0xF0FC
TMCC1_KEEP_ALIVE_COMMAND: int = 0xF0FD
TMCC1_RESYNC_COMMAND: int = 0xF0FE
TMCC1_SYNCED_COMMAND: int = 0xF0FF


@unique
class TMCC1SyncCommandEnum(TMCC1Enum):
    DISCONNECT = SyncCommandDef(TMCC1_DISCONNECT_COMMAND)
    KEEP_ALIVE = SyncCommandDef(TMCC1_KEEP_ALIVE_COMMAND)
    QUIT = SyncCommandDef(TMCC1_QUIT_COMMAND)
    REBOOT = SyncCommandDef(TMCC1_REBOOT_COMMAND, node_scope=True)
    REGISTER = SyncCommandDef(TMCC1_REGISTER_COMMAND)
    RESTART = SyncCommandDef(TMCC1_RESTART_COMMAND)
    RESYNC = SyncCommandDef(TMCC1_RESYNC_COMMAND)
    SENDING_STATE = SyncCommandDef(TMCC1_SENDING_STATE_COMMAND)
    SHUTDOWN = SyncCommandDef(TMCC1_SHUTDOWN_COMMAND, node_scope=True)
    SYNCHRONIZED = SyncCommandDef(TMCC1_SYNCED_COMMAND)
    SYNCHRONIZING = SyncCommandDef(TMCC1_SYNCING_COMMAND)
    SYNC_BEGIN = SyncCommandDef(TMCC1_SYNC_BEGIN_COMMAND)
    SYNC_COMPLETE = SyncCommandDef(TMCC1_SYNC_COMPLETE_COMMAND)
    SYNC_REQUEST = SyncCommandDef(TMCC1_SYNC_REQUEST_COMMAND)
    UPDATE = SyncCommandDef(TMCC1_UPDATE_COMMAND)
    UPGRADE = SyncCommandDef(TMCC1_UPGRADE_COMMAND)


TMCC1_HALT_COMMAND: int = 0xFFFF


@unique
class TMCC1HaltCommandEnum(TMCC1Enum):
    HALT = TMCC1CommandDef(TMCC1_HALT_COMMAND, TMCC1CommandIdentifier.HALT, is_addressable=False, filtered=True)


TMCC1_ROUTE_COMMAND: int = 0xD01F


@unique
class TMCC1RouteCommandEnum(TMCC1Enum):
    FIRE = TMCC1CommandDef(TMCC1_ROUTE_COMMAND, TMCC1CommandIdentifier.ROUTE, num_address_bits=5)


TMCC1_SWITCH_THRU_COMMAND: int = 0x4000
TMCC1_SWITCH_OUT_COMMAND: int = 0x401F
TMCC1_SWITCH_SET_ADDRESS_COMMAND: int = 0x402B


@unique
class TMCC1SwitchCommandEnum(TMCC1Enum):
    THRU = TMCC1CommandDef(TMCC1_SWITCH_THRU_COMMAND, TMCC1CommandIdentifier.SWITCH)
    OUT = TMCC1CommandDef(TMCC1_SWITCH_OUT_COMMAND, TMCC1CommandIdentifier.SWITCH)
    SET_ADDRESS = TMCC1CommandDef(TMCC1_SWITCH_SET_ADDRESS_COMMAND, TMCC1CommandIdentifier.SWITCH)


TMCC1_ACC_ON_COMMAND: int = 0x802F
TMCC1_ACC_OFF_COMMAND: int = 0x8020
TMCC1_ACC_NUMERIC_COMMAND: int = 0x8010
TMCC1_ACC_SET_ADDRESS_COMMAND: int = 0x802B

TMCC1_ACC_AUX_1_OFF_COMMAND: int = 0x8008
TMCC1_ACC_AUX_1_OPTION_1_COMMAND: int = 0x8009  # Cab1 Aux1 button
TMCC1_ACC_AUX_1_OPTION_2_COMMAND: int = 0x800A
TMCC1_ACC_AUX_1_ON_COMMAND: int = 0x800B

TMCC1_ACC_AUX_2_OFF_COMMAND: int = 0x800C
TMCC1_ACC_AUX_2_OPTION_1_COMMAND: int = 0x800D  # Cab1 Aux2 button
TMCC1_ACC_AUX_2_OPTION_2_COMMAND: int = 0x800E
TMCC1_ACC_AUX_2_ON_COMMAND: int = 0x800F

TMCC1_ACC_FRONT_COUPLER_COMMAND: int = 0x8005
TMCC1_ACC_REAR_COUPLER_COMMAND: int = 0x8006

TMCC1_ACC_BOOST_COMMAND: int = 0x8004
TMCC1_ACC_BRAKE_COMMAND: int = 0x8007
TMCC1_ACC_RELATIVE_SPEED_COMMAND: int = 0x8040


@unique
class TMCC1AuxCommandEnum(TMCC1Enum):
    SET_ADDRESS = TMCC1CommandDef(TMCC1_ACC_SET_ADDRESS_COMMAND, TMCC1CommandIdentifier.ACC)
    NUMERIC = TMCC1CommandDef(TMCC1_ACC_NUMERIC_COMMAND, TMCC1CommandIdentifier.ACC, d_max=9)
    AUX1_OFF = TMCC1CommandDef(TMCC1_ACC_AUX_1_OFF_COMMAND, TMCC1CommandIdentifier.ACC)
    AUX1_ON = TMCC1CommandDef(TMCC1_ACC_AUX_1_ON_COMMAND, TMCC1CommandIdentifier.ACC)
    AUX1_OPT_ONE = TMCC1CommandDef(TMCC1_ACC_AUX_1_OPTION_1_COMMAND, TMCC1CommandIdentifier.ACC)
    AUX1_OPT_TWO = TMCC1CommandDef(TMCC1_ACC_AUX_1_OPTION_2_COMMAND, TMCC1CommandIdentifier.ACC)
    AUX2_OFF = TMCC1CommandDef(TMCC1_ACC_AUX_2_OFF_COMMAND, TMCC1CommandIdentifier.ACC)
    AUX2_ON = TMCC1CommandDef(TMCC1_ACC_AUX_2_ON_COMMAND, TMCC1CommandIdentifier.ACC)
    AUX2_OPT_ONE = TMCC1CommandDef(TMCC1_ACC_AUX_2_OPTION_1_COMMAND, TMCC1CommandIdentifier.ACC)
    AUX2_OPT_TWO = TMCC1CommandDef(TMCC1_ACC_AUX_2_OPTION_2_COMMAND, TMCC1CommandIdentifier.ACC)
    FRONT_COUPLER = TMCC1CommandDef(TMCC1_ACC_FRONT_COUPLER_COMMAND, TMCC1CommandIdentifier.ACC)
    REAR_COUPLER = TMCC1CommandDef(TMCC1_ACC_REAR_COUPLER_COMMAND, TMCC1CommandIdentifier.ACC)
    BOOST = TMCC1CommandDef(TMCC1_ACC_BOOST_COMMAND, TMCC1CommandIdentifier.ACC)
    BRAKE = TMCC1CommandDef(TMCC1_ACC_BRAKE_COMMAND, TMCC1CommandIdentifier.ACC)
    RELATIVE_SPEED = TMCC1CommandDef(
        TMCC1_ACC_RELATIVE_SPEED_COMMAND, TMCC1CommandIdentifier.ACC, d_map=RELATIVE_SPEED_MAP
    )


# Engine/Train commands
TMCC1_TRAIN_COMMAND_MODIFIER: int = 0xC800  # Logically OR with engine command to make train command
TMCC1_TRAIN_COMMAND_PURIFIER: int = 0x07FF  # Logically AND with engine command to reset engine bits
TMCC1_ENG_ABSOLUTE_SPEED_COMMAND: int = 0x0060  # Absolute speed 0-31 encoded in last 5-bits
TMCC1_ENG_RELATIVE_SPEED_COMMAND: int = 0x0040  # Relative Speed -5 - 5 encoded in last 4-bits (offset by 5)
TMCC1_ENG_FORWARD_DIRECTION_COMMAND: int = 0x0000
TMCC1_ENG_TOGGLE_DIRECTION_COMMAND: int = 0x0001
TMCC1_ENG_REVERSE_DIRECTION_COMMAND: int = 0x0003
TMCC1_ENG_BOOST_SPEED_COMMAND: int = 0x0004
TMCC1_ENG_BRAKE_SPEED_COMMAND: int = 0x0007
TMCC1_ENG_OPEN_FRONT_COUPLER_COMMAND: int = 0x0005
TMCC1_ENG_OPEN_REAR_COUPLER_COMMAND: int = 0x0006
TMCC1_ENG_BLOW_HORN_ONE_COMMAND: int = 0x001C
TMCC1_ENG_RING_BELL_COMMAND: int = 0x001D
TMCC1_ENG_LET_OFF_SOUND_COMMAND: int = 0x001E
TMCC1_ENG_BLOW_HORN_TWO_COMMAND: int = 0x001F

TMCC1_ENG_AUX1_OFF_COMMAND: int = 0x0008
TMCC1_ENG_AUX1_OPTION_ONE_COMMAND: int = 0x0009  # Aux 1 button
TMCC1_ENG_AUX1_OPTION_TWO_COMMAND: int = 0x000A
TMCC1_ENG_AUX1_ON_COMMAND: int = 0x000B

TMCC1_ENG_AUX2_OFF_COMMAND: int = 0x000C
TMCC1_ENG_AUX2_OPTION_ONE_COMMAND: int = 0x000D  # Aux 2 button
TMCC1_ENG_AUX2_OPTION_TWO_COMMAND: int = 0x000E
TMCC1_ENG_AUX2_ON_COMMAND: int = 0x000F

TMCC1_ENG_AUX3_OPTION_ONE_COMMAND: int = 0x003B  # Aux 3 button

TMCC1_ENG_SET_MOMENTUM_LOW_COMMAND: int = 0x0028
TMCC1_ENG_SET_MOMENTUM_MEDIUM_COMMAND: int = 0x0029
TMCC1_ENG_SET_MOMENTUM_HIGH_COMMAND: int = 0x002A

TMCC1_ENG_NUMERIC_COMMAND: int = 0x0010

TMCC1_ENG_VOLUME_UP_COMMAND: int = 0x0011
TMCC1_ENG_SOUND_ONE_COMMAND: int = 0x0012
TMCC1_ENG_RPM_UP_COMMAND: int = 0x0013
TMCC1_ENG_VOLUME_DOWN_COMMAND: int = 0x0014
TMCC1_ENG_SHUTDOWN_COMMAND: int = 0x0015
TMCC1_ENG_RPM_DOWN_COMMAND: int = 0x0016
TMCC1_ENG_SOUND_TWO_COMMAND: int = 0x0017
TMCC1_ENG_FUNC_MINUS_COMMAND: int = 0x0018
TMCC1_ENG_FUNC_PLUS_COMMAND: int = 0x0019

TMCC1_ENG_SET_ADDRESS_COMMAND: int = 0x002B

TMCC1_ENG_ASSIGN_TRAIN_COMMAND: int = 0x0030
TMCC1_ENG_ASSIGN_SINGLE_FORWARD_COMMAND: int = 0x0020
TMCC1_ENG_ASSIGN_SINGLE_REVERSE_COMMAND: int = 0x0024
TMCC1_ENG_ASSIGN_HEAD_FORWARD_COMMAND: int = 0x0021
TMCC1_ENG_ASSIGN_HEAD_REVERSE_COMMAND: int = 0x0025
TMCC1_ENG_ASSIGN_MIDDLE_FORWARD_COMMAND: int = 0x0022
TMCC1_ENG_ASSIGN_MIDDLE_REVERSE_COMMAND: int = 0x0026
TMCC1_ENG_ASSIGN_REAR_FORWARD_COMMAND: int = 0x0023
TMCC1_ENG_ASSIGN_REAR_REVERSE_COMMAND: int = 0x0027
TMCC1_ENG_CLEAR_CONSIST_COMMAND: int = 0x002C

TMCC1_ROLL_SPEED: int = 1  # express speeds as simple integers
TMCC1_RESTRICTED_SPEED: int = 5
TMCC1_SLOW_SPEED: int = 10
TMCC1_MEDIUM_SPEED: int = 15
TMCC1_LIMITED_SPEED: int = 20
TMCC1_NORMAL_SPEED: int = 25
TMCC1_HIGHBALL_SPEED: int = 27

TMCC1_SPEED_MAP: dict[str, int] = {
    "STOP_HOLD": 0,
    "SH": 0,
    "STOP": 0,
    "ROLL": TMCC1_ROLL_SPEED,
    "RO": TMCC1_ROLL_SPEED,
    "RESTRICTED": TMCC1_RESTRICTED_SPEED,
    "RE": TMCC1_RESTRICTED_SPEED,
    "SLOW": TMCC1_SLOW_SPEED,
    "SL": TMCC1_SLOW_SPEED,
    "MEDIUM": TMCC1_MEDIUM_SPEED,
    "ME": TMCC1_MEDIUM_SPEED,
    "LIMITED": TMCC1_LIMITED_SPEED,
    "LI": TMCC1_LIMITED_SPEED,
    "NORMAL": TMCC1_NORMAL_SPEED,
    "NO": TMCC1_NORMAL_SPEED,
    "HIGH": TMCC1_HIGHBALL_SPEED,
    "HIGHBALL": TMCC1_HIGHBALL_SPEED,
    "HI": TMCC1_HIGHBALL_SPEED,
}

TMCC1_NUMERIC_SPEED_TO_DIRECTIVE_MAP = {s: p for p, s in TMCC1_SPEED_MAP.items() if len(p) > 2 and p != "HIGH"}


@unique
class TMCC1RRSpeedsEnum(OfficialRRSpeeds):
    STOP_HOLD = range(0, TMCC1_ROLL_SPEED)
    ROLL = range(TMCC1_ROLL_SPEED, TMCC1_RESTRICTED_SPEED)
    RESTRICTED = range(TMCC1_RESTRICTED_SPEED, TMCC1_SLOW_SPEED)
    SLOW = range(TMCC1_SLOW_SPEED, TMCC1_MEDIUM_SPEED)
    MEDIUM = range(TMCC1_MEDIUM_SPEED, TMCC1_LIMITED_SPEED)
    LIMITED = range(TMCC1_LIMITED_SPEED, TMCC1_NORMAL_SPEED)
    NORMAL = range(TMCC1_NORMAL_SPEED, TMCC1_HIGHBALL_SPEED)
    HIGHBALL = range(TMCC1_HIGHBALL_SPEED, 32)


@unique
class TMCC1EngineCommandEnum(TMCC1Enum):
    ABSOLUTE_SPEED = TMCC1CommandDef(TMCC1_ENG_ABSOLUTE_SPEED_COMMAND, d_max=31, filtered=True)
    ASSIGN_TRAIN = TMCC1CommandDef(TMCC1_ENG_ASSIGN_TRAIN_COMMAND, d_max=10)
    AUX1_OFF = TMCC1CommandDef(TMCC1_ENG_AUX1_OFF_COMMAND)
    AUX1_ON = TMCC1CommandDef(TMCC1_ENG_AUX1_ON_COMMAND)
    AUX1_OPTION_ONE = TMCC1CommandDef(TMCC1_ENG_AUX1_OPTION_ONE_COMMAND)
    AUX1_OPTION_TWO = TMCC1CommandDef(TMCC1_ENG_AUX1_OPTION_TWO_COMMAND)
    AUX2_OFF = TMCC1CommandDef(TMCC1_ENG_AUX2_OFF_COMMAND)
    AUX2_ON = TMCC1CommandDef(TMCC1_ENG_AUX2_ON_COMMAND)
    AUX2_OPTION_ONE = TMCC1CommandDef(TMCC1_ENG_AUX2_OPTION_ONE_COMMAND)
    AUX2_OPTION_TWO = TMCC1CommandDef(TMCC1_ENG_AUX2_OPTION_TWO_COMMAND)
    AUX3_OPTION_ONE = TMCC1CommandDef(TMCC1_ENG_AUX3_OPTION_ONE_COMMAND)
    BLOW_HORN_ONE = TMCC1CommandDef(TMCC1_ENG_BLOW_HORN_ONE_COMMAND, interval=100)
    BLOW_HORN_TWO = TMCC1CommandDef(TMCC1_ENG_BLOW_HORN_TWO_COMMAND, interval=100)
    BOOST_SPEED = TMCC1CommandDef(TMCC1_ENG_BOOST_SPEED_COMMAND, interval=200)
    BRAKE_SPEED = TMCC1CommandDef(TMCC1_ENG_BRAKE_SPEED_COMMAND, interval=200)
    CLEAR_CONSIST = TMCC1CommandDef(TMCC1_ENG_CLEAR_CONSIST_COMMAND)
    FORWARD_DIRECTION = TMCC1CommandDef(TMCC1_ENG_FORWARD_DIRECTION_COMMAND, filtered=False)
    FRONT_COUPLER = TMCC1CommandDef(TMCC1_ENG_OPEN_FRONT_COUPLER_COMMAND)
    FUNC_MINUS = TMCC1CommandDef(TMCC1_ENG_FUNC_MINUS_COMMAND, alias="NUMERIC", data=8)
    FUNC_PLUS = TMCC1CommandDef(TMCC1_ENG_FUNC_PLUS_COMMAND, alias="NUMERIC", data=9)
    HEAD_FORWARD = TMCC1CommandDef(TMCC1_ENG_ASSIGN_HEAD_FORWARD_COMMAND)
    HEAD_REVERSE = TMCC1CommandDef(TMCC1_ENG_ASSIGN_HEAD_REVERSE_COMMAND)
    LET_OFF = TMCC1CommandDef(TMCC1_ENG_LET_OFF_SOUND_COMMAND)
    MIDDLE_FORWARD = TMCC1CommandDef(TMCC1_ENG_ASSIGN_MIDDLE_FORWARD_COMMAND)
    MIDDLE_REVERSE = TMCC1CommandDef(TMCC1_ENG_ASSIGN_MIDDLE_REVERSE_COMMAND)
    MOMENTUM_HIGH = TMCC1CommandDef(TMCC1_ENG_SET_MOMENTUM_HIGH_COMMAND)
    MOMENTUM_LOW = TMCC1CommandDef(TMCC1_ENG_SET_MOMENTUM_LOW_COMMAND)
    MOMENTUM_MEDIUM = TMCC1CommandDef(TMCC1_ENG_SET_MOMENTUM_MEDIUM_COMMAND)
    NUMERIC = TMCC1CommandDef(TMCC1_ENG_NUMERIC_COMMAND, d_max=9)
    REAR_COUPLER = TMCC1CommandDef(TMCC1_ENG_OPEN_REAR_COUPLER_COMMAND)
    REAR_FORWARD = TMCC1CommandDef(TMCC1_ENG_ASSIGN_REAR_FORWARD_COMMAND)
    REAR_REVERSE = TMCC1CommandDef(TMCC1_ENG_ASSIGN_REAR_REVERSE_COMMAND)
    RELATIVE_SPEED = TMCC1CommandDef(TMCC1_ENG_RELATIVE_SPEED_COMMAND, d_map=RELATIVE_SPEED_MAP)
    RESET = TMCC1CommandDef(TMCC1_ENG_NUMERIC_COMMAND | 0, alias="NUMERIC", data=0)
    REVERSE_DIRECTION = TMCC1CommandDef(TMCC1_ENG_REVERSE_DIRECTION_COMMAND, filtered=False)
    RING_BELL = TMCC1CommandDef(TMCC1_ENG_RING_BELL_COMMAND)
    RPM_DOWN = TMCC1CommandDef(TMCC1_ENG_RPM_DOWN_COMMAND, alias="NUMERIC", data=6)
    RPM_UP = TMCC1CommandDef(TMCC1_ENG_RPM_UP_COMMAND, alias="NUMERIC", data=3)
    SET_ADDRESS = TMCC1CommandDef(TMCC1_ENG_SET_ADDRESS_COMMAND)
    SHUTDOWN_IMMEDIATE = TMCC1CommandDef(TMCC1_ENG_NUMERIC_COMMAND | 5, alias="NUMERIC", data=5, aux1=True)
    SINGLE_FORWARD = TMCC1CommandDef(TMCC1_ENG_ASSIGN_SINGLE_FORWARD_COMMAND)
    SINGLE_REVERSE = TMCC1CommandDef(TMCC1_ENG_ASSIGN_SINGLE_REVERSE_COMMAND)
    SMOKE_OFF = TMCC1CommandDef(TMCC1_ENG_NUMERIC_COMMAND | 8, alias="NUMERIC", data=8)
    SMOKE_ON = TMCC1CommandDef(TMCC1_ENG_NUMERIC_COMMAND | 9, alias="NUMERIC", data=9)
    SOUND_ONE = TMCC1CommandDef(TMCC1_ENG_SOUND_ONE_COMMAND, alias="NUMERIC", data=2)
    TOWER_CHATTER = TMCC1CommandDef(TMCC1_ENG_SOUND_TWO_COMMAND, alias="NUMERIC", data=7)
    START_UP_IMMEDIATE = TMCC1CommandDef(TMCC1_ENG_NUMERIC_COMMAND | 3, alias="NUMERIC", data=3, aux1=True)
    STOP_IMMEDIATE = TMCC1CommandDef(TMCC1_ENG_ABSOLUTE_SPEED_COMMAND, alias="ABSOLUTE_SPEED", data=0, filtered=True)
    TOGGLE_DIRECTION = TMCC1CommandDef(TMCC1_ENG_TOGGLE_DIRECTION_COMMAND, filtered=True)
    VOLUME_DOWN = TMCC1CommandDef(TMCC1_ENG_VOLUME_DOWN_COMMAND, alias="NUMERIC", data=4)
    VOLUME_UP = TMCC1CommandDef(TMCC1_ENG_VOLUME_UP_COMMAND, alias="NUMERIC", data=1)
    SPEED_HIGHBALL = TMCC1CommandDef(
        TMCC1_ENG_ABSOLUTE_SPEED_COMMAND | TMCC1_HIGHBALL_SPEED,
        alias="ABSOLUTE_SPEED",
        data=TMCC1_HIGHBALL_SPEED,
        filtered=True,
    )
    SPEED_LIMITED = TMCC1CommandDef(
        TMCC1_ENG_ABSOLUTE_SPEED_COMMAND | TMCC1_LIMITED_SPEED,
        alias="ABSOLUTE_SPEED",
        data=TMCC1_LIMITED_SPEED,
        filtered=True,
    )
    SPEED_MEDIUM = TMCC1CommandDef(
        TMCC1_ENG_ABSOLUTE_SPEED_COMMAND | TMCC1_MEDIUM_SPEED,
        alias="ABSOLUTE_SPEED",
        data=TMCC1_MEDIUM_SPEED,
        filtered=True,
    )
    SPEED_NORMAL = TMCC1CommandDef(
        TMCC1_ENG_ABSOLUTE_SPEED_COMMAND | TMCC1_NORMAL_SPEED,
        alias="ABSOLUTE_SPEED",
        data=TMCC1_NORMAL_SPEED,
        filtered=True,
    )
    SPEED_RESTRICTED = TMCC1CommandDef(
        TMCC1_ENG_ABSOLUTE_SPEED_COMMAND | TMCC1_RESTRICTED_SPEED,
        alias="ABSOLUTE_SPEED",
        data=TMCC1_RESTRICTED_SPEED,
        filtered=True,
    )
    SPEED_ROLL = TMCC1CommandDef(
        TMCC1_ENG_ABSOLUTE_SPEED_COMMAND | TMCC1_ROLL_SPEED,
        alias="ABSOLUTE_SPEED",
        data=TMCC1_ROLL_SPEED,
        filtered=True,
    )
    SPEED_SLOW = TMCC1CommandDef(
        TMCC1_ENG_ABSOLUTE_SPEED_COMMAND | TMCC1_SLOW_SPEED,
        alias="ABSOLUTE_SPEED",
        data=TMCC1_SLOW_SPEED,
        filtered=True,
    )
    SPEED_STOP_HOLD = TMCC1CommandDef(
        TMCC1_ENG_ABSOLUTE_SPEED_COMMAND,
        alias="ABSOLUTE_SPEED",
        data=0,
        filtered=True,
    )


TMCC1_COMMAND_TO_ALIAS_MAP = {}
for tmcc1_enum in [TMCC1EngineCommandEnum]:
    for enum in tmcc1_enum:
        if enum.is_alias:
            TMCC1_COMMAND_TO_ALIAS_MAP[enum.alias] = enum



import time
from datetime import datetime

from .constants import PdiCommand, D4Action, PDI_SOP, PDI_EOP
from ..db.comp_data import BASE_MEMORY_ENGINE_READ_MAP, CompData, CompDataMixin, CompDataHandler
from .pdi_req import PdiReq
from ..protocol.constants import CommandScope

LIONEL_EPOCH: int = 1577836800  # Midnight, Jan 1 2020 UTC


class D4Req(PdiReq, CompDataMixin):
    @staticmethod
    def lionel_timestamp(as_bytes: bool = True) -> int | bytes:
        lts = 0xFFFFFFFF & int(time.time() - LIONEL_EPOCH)
        if as_bytes is True:
            return lts.to_bytes(4, byteorder="little")
        else:
            return lts

    def __init__(
        self,
        data: bytes | int = 0,
        pdi_command: PdiCommand = PdiCommand.D4_ENGINE,
        action: D4Action = D4Action.QUERY,
        error: bool = False,
        tmcc_id: int = 0,
        post_action: int = 0,
        start: int = 0,
        data_length: int = 1,
        data_bytes: int | str | bytes | None = None,
        count: int | None = None,
        timestamp: int | None = None,
        state: CompDataMixin = None,
    ) -> None:
        super().__init__(data, pdi_command)
        self._scope = CommandScope.TRAIN if self.pdi_command == PdiCommand.D4_TRAIN else CommandScope.ENGINE
        self._record_no = self._next_record_no = self._tmcc_id = self._count = self._post_action = self._suffix = None
        self._data_length = self._data_bytes = self._start = self._timestamp = None
        self._error = error
        self._tmcc_id = tmcc_id
        if isinstance(data, bytes):
            data_len = len(self._data)
            self._record_no = int.from_bytes(self._data[1:3], byteorder="little") if data_len > 2 else None
            self._action = D4Action(self._data[3]) if data_len > 3 else None
            self._post_action = int.from_bytes(self._data[4:6]) if data_len > 5 else None
            self._suffix = int.from_bytes(self._data[8:10], byteorder="little") if data_len > 9 else None
            if self._action in {D4Action.QUERY, D4Action.UPDATE}:
                self._start = self._data[6] if data_len > 6 else None
                self._data_length = self._data[7] if data_len > 7 else None
                self._timestamp = int.from_bytes(self._data[8:12], byteorder="little") if data_len > 11 else None
                data_bytes = self._data[12:] if data_len > 12 else None
                if isinstance(data_bytes, bytes):
                    self._data_bytes = data_bytes
                    if self.start == 0 and self.data_length == PdiReq.scope_record_length(CommandScope.ENGINE):
                        self._comp_data = CompData.from_bytes(data_bytes, self.scope)
                        self._tmcc_id = self._comp_data.tmcc_id
                        self._comp_data_record = True  # mark this req as containing a complete CompData record
                    elif isinstance(data_bytes, str):
                        self._data_bytes = data_bytes[0:data_length].encode("ascii")
                        if len(data_bytes) < data_length:
                            self._data_bytes += bytes() * (data_length - len(data_bytes))
                    elif isinstance(data_bytes, int):
                        self._data_bytes = data_bytes.to_bytes(data_length, byteorder="little")
            elif self.action == D4Action.COUNT:
                self._scope = CommandScope.BASE
                self._count = int.from_bytes(self._data[6:8], byteorder="little") if data_len > 7 else None
            elif self.action == D4Action.MAP:
                self._suffix = None
                if data_len > 9:
                    addr_str = ""
                    for i in range(6, 10):
                        addr_str += chr(self._data[i])
                    self._tmcc_id = int(addr_str)
            elif self.action in {D4Action.FIRST_REC, D4Action.NEXT_REC}:
                if self._action == D4Action.NEXT_REC:
                    self._post_action = int.from_bytes(self._data[4:6]) if data_len > 5 else None
                    self._start = 0
                    self._data_length = self._data[7] if data_len > 7 else None
                    self._next_record_no = (
                        int.from_bytes(self._data[8:10], byteorder="little") if data_len > 9 else None
                    )
                else:
                    self._next_record_no = self.record_no
                    self._scope = CommandScope.BASE  # send first record information to Base
        else:
            from ..db.engine_state import EngineState

            self._action = action
            self._record_no = int(data)
            self._post_action = post_action
            self._start = start
            self._data_length = data_length
            self._data_bytes = data_bytes
            self._count = count
            self._timestamp = timestamp
            self._state = state
            if state and isinstance(state, EngineState):
                self._tmcc_id = state.tmcc_id
                self._start = 0
                self._data_length = PdiReq.scope_record_length(CommandScope.ENGINE)
                if isinstance(state, CompDataMixin):
                    self._data_bytes = state.comp_data.as_bytes()
            elif self.action in {D4Action.FIRST_REC, D4Action.COUNT}:
                self._scope = CommandScope.BASE

    @property
    def as_key(self):
        return self.record_no, self.pdi_command, self.action, self.scope

    @property
    def record_no(self) -> int:
        return self._record_no

    @property
    def action(self) -> D4Action:
        return self._action

    @property
    def post_action(self) -> int:
        return self._post_action

    @property
    def suffix(self) -> int:
        return self._suffix

    @property
    def start(self) -> int:
        return self._start

    @property
    def data_length(self) -> int:
        return self._data_length

    @property
    def count(self) -> int:
        return self._count

    @property
    def next_record_no(self) -> int:
        return self._next_record_no

    @property
    def timestamp(self) -> int:
        return self._timestamp

    @property
    def name(self) -> str | None:
        if self.comp_data:
            return self.comp_data.road_name
        else:
            return None

    @property
    def number(self) -> str | None:
        if self.comp_data:
            return self.comp_data.road_number
        else:
            return None

    @property
    def timestamp_str(self) -> str:
        if self.timestamp:
            return datetime.fromtimestamp(self.timestamp + LIONEL_EPOCH).strftime("%Y-%m-%d %H:%M:%S")
        return ""

    @property
    def payload(self) -> str:
        if self.action:
            ct = tmcc = dl = di = db = ts = ""
            op = self.action.title
            rn = f" #{self.record_no}" if self.record_no is not None else ""
            sf = f" {self.suffix}" if self.suffix is not None else ""
            if self.action == D4Action.COUNT:
                rn = ""
                ct = f": {self.count}" if self.count is not None else ""
            elif self.action == D4Action.MAP:
                tmcc = f" TMCC ID: {self.tmcc_id}" if self.tmcc_id else ""
                if self.record_no == 0xFFFF:
                    rn = " Not Found"
            elif self.action in {D4Action.QUERY, D4Action.UPDATE}:
                tmcc = f" TMCC ID: {self.tmcc_id}" if self.tmcc_id else ""
                sf = ""
                ts = f" {self.timestamp_str}"
                di = f" Index: {hex(self.start)}" if self.start is not None else ""
                dl = f" Length: {self.data_length}" if self.data_length is not None else ""
                tpl = BASE_MEMORY_ENGINE_READ_MAP.get(self.start, None)
                if isinstance(tpl, CompDataHandler) and self._data_bytes and (self.data_length == tpl.length):
                    db = f" Data: {tpl.from_bytes(self._data_bytes)}"
                else:
                    db = (
                        f" Data: {self._data_bytes.hex()}"
                        if self._data_bytes is not None and len(self._data_bytes) < 0xC0
                        else " Data: None"
                    )
            return f"{op}{tmcc}{rn}{ct}{sf}{di}{dl}{db}{ts} ({self.packet})"
        return super().payload

    @property
    def as_bytes(self) -> bytes:
        if self._original:
            return self._original
        byte_str = self.pdi_command.as_bytes
        byte_str += self.record_no.to_bytes(2, byteorder="little")
        byte_str += self.action.as_bytes
        if self.action in {D4Action.COUNT, D4Action.MAP, D4Action.NEXT_REC, D4Action.QUERY, D4Action.UPDATE}:
            byte_str += (self.post_action if self.post_action else 0).to_bytes(2, byteorder="little")
        if self.action == D4Action.COUNT:
            byte_str += self.count.to_bytes(2, byteorder="little") if self.count is not None else bytes()
            byte_str += self.suffix.to_bytes(2, byteorder="little") if self.suffix is not None else bytes()
        elif self.action == D4Action.FIRST_REC:
            self._scope = CommandScope.SYSTEM
            byte_str += (0).to_bytes(1, byteorder="big")
        elif self.action == D4Action.NEXT_REC:
            self._scope = CommandScope.SYSTEM
            byte_str += self.start.to_bytes(1, byteorder="big") if self.start is not None else bytes()
            byte_str += self.data_length.to_bytes(1, byteorder="big") if self.data_length is not None else bytes()
            byte_str += (
                self.next_record_no.to_bytes(2, byteorder="little") if self.next_record_no is not None else bytes()
            )
        elif self.action == D4Action.MAP:
            byte_str += str(self.tmcc_id).zfill(4).encode("ascii") if self.tmcc_id else bytes()
        elif self.action in {D4Action.QUERY, D4Action.UPDATE}:
            byte_str += self.start.to_bytes(1, byteorder="big") if self.start is not None else bytes()
            byte_str += self.data_length.to_bytes(1, byteorder="big") if self.data_length is not None else bytes()
            byte_str += (
                self.timestamp.to_bytes(4, byteorder="little")
                if self.timestamp is not None
                else self.lionel_timestamp()
            )
            if self._data_bytes is not None:
                byte_str += self._data_bytes
        byte_str, checksum = self._calculate_checksum(byte_str)
        byte_str = PDI_SOP.to_bytes(1, byteorder="big") + byte_str
        byte_str += checksum
        byte_str += PDI_EOP.to_bytes(1, byteorder="big")
        return byte_str



from __future__ import annotations

import logging
import select
import socket
import threading
import time

# from signal import signal, SIGPIPE, SIG_DFL
from threading import Condition, Thread

from .base_req import BaseReq
from .constants import KEEP_ALIVE_CMD, PDI_SOP, PdiCommand, TMCC_TX, TMCC4_TX
from .pdi_listener import PdiListener
from .pdi_req import PdiReq, TmccReq
from ..protocol.command_req import CommandReq
from ..protocol.constants import (
    DEFAULT_BASE_PORT,
    DEFAULT_QUEUE_SIZE,
    CommandScope,
    PROGRAM_NAME,
)
from ..utils.pollable_queue import PollableQueue

log = logging.getLogger(__name__)


class Base3Buffer(Thread):
    # noinspection GrazieInspection
    """
    Send and receive PDI command packets to/from a Lionel Base 3 or LCS WiFi module.
    """

    _instance: None = None
    _lock = threading.RLock()

    @classmethod
    def stop(cls) -> None:
        if cls._instance is not None:
            cls._instance.shutdown()
            cls._instance = None

    @classmethod
    def get(cls) -> Base3Buffer:
        if cls._instance is None:
            raise AttributeError("Base3Buffer has not been initialized")
        return cls._instance

    @classmethod
    def base_address(cls) -> str:
        if cls._instance is None:
            raise AttributeError("Base3Buffer has not been initialized")
        # noinspection PyProtectedMember
        return cls._instance._base3_addr

    @classmethod
    def request_state_update(cls, tmcc_id: int, scope: CommandScope) -> None:
        if cls._instance is not None:
            if 1 <= tmcc_id <= 99:
                cls.enqueue_command(BaseReq(tmcc_id, PdiCommand.BASE_MEMORY, scope=scope).as_bytes)

    @classmethod
    def enqueue_command(cls, data: bytes) -> None:
        if cls._instance is not None and data:
            cls._instance.send(data)

    def __init__(
        self,
        base3_addr: str,
        base3_port: int = DEFAULT_BASE_PORT,
        buffer_size: int = DEFAULT_QUEUE_SIZE,
        listener: PdiListener = None,
    ) -> None:
        if self._initialized:
            return
        else:
            self._initialized = True
        super().__init__(daemon=True, name=f"{PROGRAM_NAME} Base3 Interface")
        # ip address and port to connect
        self._base3_addr = base3_addr
        self._base3_port = base3_port
        # data read from the Base 3 is sent to a PdiListener to decode and act on
        self._listener = listener
        self._is_running = True
        self._last_output_at = 0  # used to throttle writes to LCS Base3
        # data to send to the Base 3 is written into a queue, which is drained by the thread
        # created when this instance is started
        self._send_queue: PollableQueue[bytes] = PollableQueue(buffer_size)
        self._send_cv = Condition()
        # we must send a keepalive packet to the Base 3 every few seconds to keep it
        # from closing the connection
        self._keep_alive = KeepAlive(self)
        self.start()

    def __new__(cls, *args, **kwargs):
        """
        Provides singleton functionality. We only want one instance
        of this class in a process
        """
        with cls._lock:
            if Base3Buffer._instance is None:
                # noinspection PyTypeChecker
                Base3Buffer._instance = super(Base3Buffer, cls).__new__(cls)
                Base3Buffer._instance._initialized = False
            return Base3Buffer._instance

    def send(self, data: bytes) -> None:
        if data:
            from ..protocol.multibyte.multibyte_command_req import MultiByteReq

            # If we are sending a multibyte TMCC or TMCC_4D command, we have to break
            # it down into 3-7 byte packets; this needs to be done here so sync_state
            # in the calling layer gets a complete command
            cmd_bytes = data[2:-2]
            is_mvb, is_d4 = MultiByteReq.vet_bytes(cmd_bytes, raise_exception=False)
            if data[1] in {TMCC_TX, TMCC4_TX} and is_mvb:
                tmcc_cmd = CommandReq.from_bytes(cmd_bytes)
                # This is a legacy/tmcc2 multibyte parameter command. We have to send it
                # as 3 3 byte packets, using PdiCommand.TMCC_RX
                for packet in TmccReq.as_packets(tmcc_cmd):
                    self.send(packet)  # recursive call
                    time.sleep(0.001)
                # do a sync_state on the complete command
                self.sync_state(data)
            else:
                with self._send_cv:
                    self._send_queue.put(data)
                    self._send_cv.notify_all()

    @staticmethod
    def _current_milli_time() -> int:
        """
        Return the current time, in milliseconds past the "epoch"
        """
        return round(time.time() * 1000)

    def run(self) -> None:
        # signal(SIGPIPE, SIG_DFL)
        keep_trying = 10  # when we can't send a packet, retry 10 times
        while self._is_running and keep_trying:
            with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
                try:
                    s.connect((str(self._base3_addr), self._base3_port))
                    # we want to wait on either data being available to send to the Base3 of
                    # data available from the Base 3 to process
                    socket_list = [s, self._send_queue]
                    while self._is_running and keep_trying:
                        try:
                            readable, _, _ = select.select(socket_list, [], [])
                            for sock in readable:
                                if sock == self._send_queue:
                                    received = None
                                    sending = sock.get()
                                    # millis_since_last_output = self._current_milli_time() - self._last_output_at
                                    # if millis_since_last_output < DEFAULT_BASE_THROTTLE_DELAY:
                                    #     time.sleep((DEFAULT_BASE_THROTTLE_DELAY - millis_since_last_output) / 1000.0)
                                    s.sendall(sending.hex().upper().encode())
                                    self._last_output_at = self._current_milli_time()
                                    # update base3 with new state; required if command is a tmcc_tx
                                    try:
                                        self.sync_state(sending)
                                    except ValueError as ve:
                                        # TODO: we get exceptions here if we send a multi-byte tmcc command
                                        # TODO: because they are packet-ized into 3-byte chunks that sync-state
                                        # TODO cannot yet handle
                                        log.debug(ve)
                                    except Exception as e:
                                        log.exception(e)
                                else:
                                    sending = None
                                    received = bytes.fromhex(s.recv(512).decode(errors="ignore"))
                                    # but there is more trickiness; The Base 3 sends ascii characters
                                    # so when we receive: 'D12729DF', this actually is sent as eight
                                    # characters; D, 1, 2, 7, 2, 9, D, F, so we must decode the 8
                                    # received bytes into 8 ASCII characters, then interpret that
                                    # ASCII string as Hex representation to arrive at 0xd12729df...
                                if self._listener and received:
                                    self._listener.offer(received)
                            keep_trying = 10
                        except BrokenPipeError as bpe:
                            # keep trying; unix can sometimes just hang up
                            if sending is not None:
                                log.info(f"Exception sending: 0x{sending.hex(':').upper()}; retrying ({bpe})")
                                self.send(sending)
                            elif received is not None:
                                log.info(f"Exception receiving: 0x{received.hex(':').upper()}; retrying: {bpe}")
                            else:
                                log.exception(bpe)
                            keep_trying -= 1
                            break  # continues to outer loop
                        except ValueError as ve:
                            log.debug(ve)
                except OSError as oe:
                    log.info(
                        f"No response from Lionel Base 3 at {self._base3_addr}; is the Base 3 turned on? Retrying..."
                    )
                    log.exception(oe)
                    keep_trying -= 1
                    if keep_trying <= 0:
                        raise oe
                    else:
                        time.sleep(30 if oe.errno == 113 else 1)

    def shutdown(self) -> None:
        with self._lock:
            self._is_running = False

    @classmethod
    def sync_state(cls, data: bytes, pdi_req: PdiReq = None) -> None:
        """
        Send State Update to Base 3 if it is available and if this
        command packet is relevant
        """
        if cls._instance is None or data == KEEP_ALIVE_CMD:  # if no base 3 or ping, nothing to do
            return
        if data:
            tmcc_cmds = []
            if data[0] == PDI_SOP:  # it's a Base 3 cmd, we only care about TMCC TX/TMCC4_TX
                if len(data) > 2 and data[1] in {PdiCommand.TMCC_TX, PdiCommand.TMCC4_TX}:
                    try:
                        if pdi_req is None:
                            pdi_req = PdiReq.from_bytes(data)
                        if isinstance(pdi_req, TmccReq) and pdi_req.pdi_command in {
                            PdiCommand.TMCC_TX,
                            PdiCommand.TMCC4_TX,
                        }:
                            tmcc_cmds.append(pdi_req.tmcc_command)
                    except NotImplementedError:
                        return  # ignore exceptions; most likely it's a multibyte cmd
            else:
                # convert the byte stream into one or more commands
                command_seq = bytes()
                for b in data:
                    from ..protocol.command_req import TMCC_FIRST_BYTE_TO_INTERPRETER

                    # When we sync state, states are expressed as the byte-string representations
                    # of the command(s) that generate them. This means that the byte string we
                    # received here likely consists of multiple commands. CommandReq.from_bytes
                    # expects to receive only one command's worth of bytes at a time, so we have
                    # to break up the byte stream back into the component byte strings for each cmd.
                    if b in TMCC_FIRST_BYTE_TO_INTERPRETER and len(command_seq) >= 3:
                        if command_seq:
                            try:
                                tmcc_cmds.append(CommandReq.from_bytes(command_seq))
                                command_seq = bytes()
                            except ValueError:
                                pass
                    command_seq += b.to_bytes(1, byteorder="big")
                if command_seq:
                    tmcc_cmds.append(CommandReq.from_bytes(command_seq))
            for tmcc_cmd in tmcc_cmds:
                # is it a command that requires a state sync?
                sync_reqs = BaseReq.update_eng(tmcc_cmd)
                if sync_reqs:
                    for sync_req in sync_reqs:
                        cls._instance.send(sync_req.as_bytes)


class KeepAlive(Thread):
    """
    The Base 3 needs to receive a keep-alive signal in order for the
    connection to be maintained. This class/thread sends the keep-alive
    packet every 2 seconds
    """

    def __init__(self, writer: Base3Buffer) -> None:
        super().__init__(daemon=True, name=f"{PROGRAM_NAME} Base3 Keep Alive")
        self._writer = writer
        self._is_running = True
        self.start()

    def run(self) -> None:
        while self._is_running:
            self._writer.send(KEEP_ALIVE_CMD)
            time.sleep(2.0)



from __future__ import annotations

from typing import Dict

import sys

from .multibyte_command_req import MultiByteReq
from .multibyte_constants import TMCC2MaskingControl, TMCC2ParameterEnum

if sys.version_info >= (3, 11):
    from typing import Self
elif sys.version_info >= (3, 9):
    from typing_extensions import Self

from ..constants import DEFAULT_ADDRESS, CommandScope
from ..tmcc2.tmcc2_constants import LEGACY_TRAIN_COMMAND_PREFIX
from .multibyte_constants import TMCC2ParameterIndex
from .multibyte_constants import TMCC2RailSoundsDialogControl
from .multibyte_constants import TMCC2RailSoundsEffectsControl, TMCC2EffectsControl
from .multibyte_constants import TMCC2LightingControl


# noinspection PyTypeChecker
PARAMETER_ENUM_TO_INDEX_MAP: Dict[TMCC2ParameterEnum, TMCC2ParameterIndex] = {
    TMCC2RailSoundsDialogControl: TMCC2ParameterIndex.DIALOG_TRIGGERS,
    TMCC2RailSoundsEffectsControl: TMCC2ParameterIndex.EFFECTS_TRIGGERS,
    TMCC2MaskingControl: TMCC2ParameterIndex.MASKING_CONTROLS,
    TMCC2EffectsControl: TMCC2ParameterIndex.EFFECTS_CONTROLS,
    TMCC2LightingControl: TMCC2ParameterIndex.LIGHTING_CONTROLS,
}

PARAMETER_INDEX_TO_ENUM_MAP = {s: p for p, s in PARAMETER_ENUM_TO_INDEX_MAP.items()}


class ParameterCommandReq(MultiByteReq):
    @classmethod
    def build(
        cls, command: TMCC2ParameterEnum, address: int = DEFAULT_ADDRESS, data: int = 0, scope: CommandScope = None
    ) -> Self:
        return ParameterCommandReq(command, address, data, scope)

    @classmethod
    def from_bytes(cls, param: bytes, from_tmcc_rx: bool = False, is_tmcc4: bool = False) -> Self:
        is_pc, is_d4 = cls.vet_bytes(param, "Parameter")
        if is_pc is True:
            index = 0x00FF & int.from_bytes(param[1:3], byteorder="big")
            try:
                pi = TMCC2ParameterIndex(index)
            except ValueError:
                raise ValueError(f"Invalid parameter command: : {param.hex(':')}")
            if pi in PARAMETER_INDEX_TO_ENUM_MAP:
                param_enum = PARAMETER_INDEX_TO_ENUM_MAP[pi]
                if is_d4 is False:
                    command = int(param[5])
                else:  # account for 4-digit address encoded after each 3 bytes
                    command = int(param[9])
                cmd_enum = param_enum.by_value(command)
                if cmd_enum is not None:
                    scope = cmd_enum.scope
                    if int(param[0]) == LEGACY_TRAIN_COMMAND_PREFIX:
                        scope = CommandScope.TRAIN
                    # build_req the request and return
                    data = 0
                    p_arg = param[1:3] if is_d4 is False else param[1:7]
                    address = cmd_enum.value.address_from_bytes(p_arg)
                    cmd_req = ParameterCommandReq.build(cmd_enum, address, data, scope)
                    if from_tmcc_rx is True:
                        cmd_req._is_tmcc_rx = True
                    return cmd_req
        raise ValueError(f"Invalid parameter command: : {param.hex(':')}")

    def __init__(
        self,
        command_def_enum: TMCC2ParameterEnum,
        address: int = DEFAULT_ADDRESS,
        data: int = 0,
        scope: CommandScope = None,
    ) -> None:
        super().__init__(command_def_enum, address, data, scope)

    @property
    def index_byte(self) -> bytes:
        # noinspection PyTypeChecker
        return PARAMETER_ENUM_TO_INDEX_MAP[type(self._command_def_enum)].to_bytes(1, byteorder="big")

    @property
    def data_byte(self) -> bytes:
        return self.command_def.bits.to_bytes(1, byteorder="big")
